;;; Directory Local Variables            -*- no-byte-compile: t -*-
;;; For more information see (info "(emacs) Directory Variables")

((nil . ((ispell-buffer-session-localwords . ("dir" "hypercore" "hyperdrive" "hyperdrives" "hyperdrive's" "args" "systemd" "minibuffer" "petname" "petnames" "org" "plist" "plists" "alist" "alists" "existsp" "ETag" "streamable" "DNSLink" "ewoc" "struct" "ENTRY's" "localhost" "imenu" "mtime" "accessor" "http" "prepended" "prepend" "hostname" "whitespace" "namespace" "filesystem" "hostnames" "subdirectories" "unsets" "finalizer" "subprocess" "autodetect" "transclusion" "ushin"))))
 (emacs-lisp-mode . ((eval . (display-fill-column-indicator-mode))
                     (fill-column . 80)))
 (org-mode . ((org-id-link-to-org-use-id . nil))))
