;; Generated package description from hyperdrive.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "hyperdrive" "0.60.20251120.145618" "P2P filesystem" '((emacs "28.1") (map "3.0") (compat "30.0.0.0") (org "9.7.6") (plz "0.9.1") (persist "0.8") (taxy-magit-section "0.14") (transient "0.8.0")) :commit "ba7cdf3fd361b6d79583620fe0cc71421bc7b287" :authors '(("Joseph Turner" . "joseph@ushin.org")) :maintainer '("Joseph Turner" . "~ushin/ushin@lists.sr.ht") :url "https://git.sr.ht/~ushin/hyperdrive.el")
